module sphere_gen(CLK, reset, lock, init, busy, done,
                  Ax, Ay, Az, Bx, By, Bz, stepsize, minorR, majorR);
   input CLK, reset, lock, init;
   wire  CLK, reset, lock, init;
   output       busy, done;
   reg          busy, done;
   output [15:0] Ax, Ay, Az, Bx, By, Bz;
   wire [15:0]   Ax, Ay, Az, Bx, By, Bz;

   reg [15:0]    prev_x, prev_y, prev_z;
   wire [15:0]   next_x, next_y, next_z;

   reg [15:0]    row_initx[0:127];
   reg [15:0]    row_inity[0:127];
   reg [15:0]    row_initz[0:127];
   reg [7:0]     init_count;

   assign        Ax[15:0] = prev_x[15:0];
   assign        Ay[15:0] = prev_y[15:0];
   assign        Az[15:0] = prev_z[15:0];
   assign        Bx[15:0] = next_x[15:0];
   assign        By[15:0] = next_y[15:0];
   assign        Bz[15:0] = next_z[15:0];
   
   input [15:0]  stepsize, minorR, majorR;
   wire [15:0]   stepsize, minorR, majorR;

   reg [3:0]     state;
   reg [15:0]    d_step;
   reg [15:0]    phi_step, theta_step;

   parameter     done_looping = 4'd0,
                   phi_loop1 = 4'd1,
                   phi_loop2 = 4'd2,
                   wait_for_init_low = 4'd3,
                   theta_loop1 = 4'd4
                   ;

   wire [15:0]   sin_phi, cos_phi, sin_theta, cos_theta;
   sineTable9x16bit sineTable_phi(CLK, phi_step[8:0], sin_phi[15:0]);
   sineTable9x16bit sineTable_theta(CLK, theta_step[8:0], sin_theta[15:0]);
   cosineTable9x16bit cosineTable_phi(CLK, phi_step[8:0], cos_phi[15:0]);
   cosineTable9x16bit cosineTable_theta(CLK, theta_step[8:0], cos_theta[15:0]);

   wire [17:0]   minorR_cos_theta, minorR_sin_theta;
   signed_mult mult_minorR_cos_theta(minorR_cos_theta[17:0],
                                     { cos_theta[15:0], 2'b00 },
                                     { minorR[15:0], 2'b00} );
   signed_mult mult_minorR_sin_theta(minorR_sin_theta[17:0],
                                     { sin_theta[15:0], 2'b00 },
                                     { minorR[15:0], 2'b00} );
   
   // Ax = cos(phi_step) * (minorR_cos_theta + majorR)
   // Ay = minorR_sin_theta
   // Az = -sin(phi_step) * (minorR_cos_theta + majorR)
   wire [17:0]   mult_Ax, mult_Az;
   signed_mult multiply_Ax(mult_Ax[17:0], { cos_phi[15:0], 2'b00 },
                       minorR_cos_theta[17:0] + { majorR[15:0] + 2'b00 });
   signed_mult multiply_Az(mult_Az[17:0], { -sin_phi[15:0], 2'b00 },
                       minorR_cos_theta[17:0] + { majorR[15:0] + 2'b00 });
   assign        next_x[15:0] = mult_Ax[17:2];
   assign        next_y[15:0] = minorR_sin_theta[17:2];
   assign        next_z[15:0] = mult_Az[17:2];
   
   always @(posedge CLK) begin
      /*
      $display("mult_Ax = %h, cos_phi = %h, majorR = %h, phi_step = %h, phi_step + 128 = %h",
               mult_Ax[17:2], cos_phi[15:0], majorR[15:0], phi_step, phi_step + 9'd128);
      $display("minorR_cos_theta = %h, cos_theta = %h, minorR = %h",
               minorR_cos_theta[17:2], cos_theta[15:0], minorR[15:0]);
      $display("minorR_sin_theta = %h, sin_theta = %h, minorR = %h",
               minorR_sin_theta[17:2], sin_theta[15:0], minorR[15:0]);
      $display("mult_Az = %h, cos_phi = %h, majorR = %h",
               mult_Az[17:2], cos_phi[15:0], majorR[15:0]);
       */
      if (reset) begin
         prev_x <= 16'd0;
         prev_y <= 16'd0;
         prev_z <= 16'd0;
  
         phi_step <= 16'd0;
         theta_step <= 16'd0;
         
         busy <= 1'b0;
         done <= 1'b0;

         state <= done_looping;
         $display("sphere_gen: being reset");
      end
      else if (lock) begin
         case (state)
           done_looping: begin
              if (init == 1'b1) begin
                 // delta step. 1, 2, 3, ...
                 d_step <= stepsize;

                 // advance prev_[xyz]
                 prev_x[15:0] <= next_x[15:0];
                 prev_y[15:0] <= next_y[15:0];
                 prev_z[15:0] <= next_z[15:0];

                 row_initx[init_count] <= next_x[15:0];
                 row_inity[init_count] <= next_y[15:0];
                 row_initz[init_count] <= next_z[15:0];
                 init_count <= init_count + 8'd1;
                 // phi = delta, theta = 0
                 phi_step <= stepsize;
                 theta_step <= 16'd0;
                 init_count <= 8'd0;
         
                 // not busy yet (no valid results out)
                 busy <= 1'b0;
                 // not done yet
                 done <= 1'b0;
                 
                 state <= phi_loop1;
                 $display("sphere_gen: in done_looping, starting a new set. xyz = (%h, %h, %h)",
                          next_x, next_y, next_z);
              end // if (init == 1'b1)
              else begin
                 //$display("sphere_gen: in done_looping, with nothing to do");
                 busy <= 1'b0;
                 done <= 1'b0;
              end
           end // case: done_looping
           phi_loop1: begin
              if (theta_step > 512) begin
                 busy <= 1'b0;
                 done <= 1'b0;
                 $display("Theta is pretty much done");
                 
                 state <= wait_for_init_low;
              end
              if (phi_step > 512) begin
                 // not done yet.
                 busy <= 1'b0;
                 done <= 1'b0;

                 // advance prev_[xyz]
                 prev_x[15:0] <= next_x[15:0];
                 prev_y[15:0] <= next_y[15:0];
                 prev_z[15:0] <= next_z[15:0];

                 row_initx[init_count] <= next_x[15:0];
                 row_inity[init_count] <= next_y[15:0];
                 row_initz[init_count] <= next_z[15:0];
                 init_count <= 8'd0;

                 // advance to next loop
                 theta_step <= theta_step + d_step;
                 phi_step <= 16'd0;

                 //state <= wait_for_init_low;
                 //state <= phi_loop2;
              end // if (phi_step > 512)
              else begin
                 if (phi_step != 16'd0) begin
                    // have valid results
                    busy <= 1'b1;
                 end
                 else begin
                    busy <= 1'b0;
                 end
                 // not done
                 done <= 1'b0;

                 // advance prev_[xyz]
                 prev_x[15:0] <= next_x[15:0];
                 prev_y[15:0] <= next_y[15:0];
                 prev_z[15:0] <= next_z[15:0];

                 // advance phi.
                 // (when debugging and removing the phi_loop2 case, do NOT increment here)
                 phi_step <= phi_step + d_step;
                 //state <= phi_loop2;
                 
                 $display("[(%h, %h, %h), (%h, %h, %h)] phi = %d, theta = %d, init_count = %d",
                          prev_x, prev_y, prev_z, next_x, next_y, next_z, phi_step, theta_step, init_count);
              end
           end // case: phi_loop1
           phi_loop2: begin
              if (theta_step != 16'd0) begin
                 busy <= 1'b1;
                 done <= 1'b0;
                 //                 $display("phi_step = %d, theta_step = %d, init_count = %d, next_xyz = (%h, %h, %h)",
                 //phi_step, theta_step, init_count, next_x, next_y, next_z);

                 prev_x[15:0] <= row_initx[init_count];
                 prev_y[15:0] <= row_inity[init_count];
                 prev_z[15:0] <= row_initz[init_count];
              end
              else begin
                // still first shot. no alternate points to demonstrate
                 busy <= 1'b0;
                 done <= 1'b0;
              end // else: !if(theta_step != 16'd0)
              
              row_initx[init_count] <= next_x[15:0];
              row_inity[init_count] <= next_y[15:0];
              row_initz[init_count] <= next_z[15:0];
              init_count <= init_count + 8'd1;

              // advance phi.
              phi_step <= phi_step + d_step;

              state <= phi_loop1;
           end
           theta_loop1: begin
              if (phi_step > 512) begin
                 // not done yet.
                 busy <= 1'b1;
                 done <= 1'b0;

                 // advance prev_[xyz]
                 prev_x[15:0] <= next_x[15:0];
                 prev_y[15:0] <= next_y[15:0];
                 prev_z[15:0] <= next_z[15:0];

                 $display("[(%h, %h, %h), (%h, %h, %h)] phi = %d, theta = %d, init_count = %d",
                          prev_x, prev_y, prev_z, next_x, next_y, next_z, phi_step, theta_step, init_count);

                 
                 state <= wait_for_init_low;

                 $display("sphere_gen: inside looping with phi_step = %h > 512", phi_step);
              end
              else begin
                 //$display("stepsize = %d, phi_step = %d, d_step = %d, [(%h, %h, %h), (%h, %h, %h)]",
                 //stepsize, phi_step, d_step, prev_x, prev_y, prev_z, next_x, next_y, next_z);

                 // initial case
                 prev_x[15:0] <= next_x[15:0];
                 prev_y[15:0] <= next_y[15:0];
                 prev_z[15:0] <= next_z[15:0];
                 
                 // turn busy on
                 busy <= 1'b1;

                 // not done
                 done <= 1'b0;
                 // advance
                 phi_step <= phi_step + d_step;
              end // else: !if(phi_step > 512)
           end // case: looping
           wait_for_init_low: begin
              // now we're done
              busy <= 1'b0;
              done <= 1'b1;

              if (~init) begin
                 state <= done_looping;
              end
           end
         endcase // case (state)
      end
   end // always @ (posedge CLK)

endmodule // sphere_gen

  
      
