// Clock parameters
`define TICK      #2
`define PHASE     #5
`define CYCLE     #10

module TOP;
   reg          MRST, CLK;
   reg [3:0]    object_state;
   reg [3:0]    comp_state;
   reg [3:0]    raster_state;
   

   // edge list
   reg signed [15:0] edge_Ax[0:255], edge_Ay[0:255], edge_Az[0:255],
       edge_Bx[0:255], edge_By[0:255], edge_Bz[0:255];
   reg [7:0]    edge_count;
   reg [7:0]    edge_iterator;
   // computed edges
   reg signed [15:0] comp_Ax[0:255], comp_Ay[0:255], comp_Az[0:255],
       comp_Bx[0:255], comp_By[0:255], comp_Bz[0:255];
   reg [7:0]    comp_count;
   reg [7:0]    comp_iterator;

   // circle
   reg          circle_init;
   wire         circle_busy, circle_done;
   wire [15:0]  circle_Ax, circle_Ay, circle_Az, circle_Bx, circle_By, circle_Bz;
   reg [15:0]   circle_stepsize;

   // bresenhem
   reg          bresenhem_init;
   wire         bresenhem_busy, bresenhem_done;
   reg [9:0]    bre_Ax, bre_Ay, bre_Bx, bre_By;

   parameter     SCREEN_TOP = 10'd1,
                 SCREEN_DOWN = 10'd238,
                 SCREEN_LEFT = 10'd3,
                 SCREEN_RIGHT = 10'd319,
                 SCREEN_XHALF = 158,
                 SCREEN_YHALF = 118;

   // Create a 50% duty cycle clock
   initial begin
      CLK = 1;
      forever begin
	       `PHASE CLK = 0;
	       `PHASE CLK = 1;
      end
   end

   bresenhem primary_bresenhem(.CLK(CLK),
                               .reset(MRST),
                               .init(bresenhem_init),
                               .lock(1'b1),
                               .busy(bresenhem_busy),
                               .done(bresenhem_done),
                               .Ax_in(bre_Ax),
                               .Ay_in(bre_Ay),
                               .Bx_in(bre_Bx),
                               .By_in(bre_By));

   circle_gen circle_generator(.CLK(CLK), .reset(MRST), .init(circle_init), .lock(1'b1),
                               .busy(circle_busy), .done(circle_done),
                               .Ax(circle_Ax), .Ay(circle_Ay), .Az(circle_Az),
                               .Bx(circle_Bx), .By(circle_By), .Bz(circle_Bz),
                               .stepsize(circle_stepsize));

   // transformation wires
   reg [55:0]  graphic_word;
   reg         translater_init, scaler_init, rotater_init;
   wire [4:0]  knob_addr;
   reg [15:0]  knob_val; // obvious
   wire [15:0] transarg_x, transarg_y, transarg_z; // non-knob-multiplied
   wire [17:0] knobmult_x, knobmult_y, knobmult_z;
   wire [15:0] transop_x, transop_y, transop_z; // processed op (knob-multiplied or not)
   wire [1:0]  rotate_axis;
   // status wires per module
   wire        translater_busy_A, translater_done_A;
   wire        translater_busy_B, translater_done_B;
   wire        scaler_busy_A, scaler_done_A;
   wire        scaler_busy_B, scaler_done_B;
   wire        rotater_busy_A, rotater_done_A;
   wire        rotater_busy_B, rotater_done_B;
   // all status wires
   wire        translater_busy, translater_done;
   wire        scaler_busy, scaler_done;
   wire        rotater_busy, rotater_done;
   // if either are busy, don't advance
   assign      translater_busy = translater_busy_A | translater_busy_B;
   assign      scaler_busy = scaler_busy_A | scaler_busy_B;
   assign      rotater_busy = rotater_busy_A | rotater_busy_B;
   // if both aren't done, not done
   assign      translater_done = translater_done_A & translater_done_B;
   assign      scaler_done = scaler_done_A & scaler_done_B;
   assign      rotater_done = rotater_done_A & rotater_done_B;
   // temporary registers
   reg [15:0]  transform_Ax, transform_Ay, transform_Az;
   reg [15:0]  transform_Bx, transform_By, transform_Bz;
   wire [15:0] next_Ax, next_Ay, next_Az;
   wire [15:0] next_Bx, next_By, next_Bz;
   // transformation and rotation selection
   wire [2:0]   transform_sel;
   assign      rotate_axis[1:0] = transform_sel[1:0];
   // transform ops
   reg [55:0]   graphic_ops [0:4];
   reg [4:0]   graphic_ops_count;
   reg [4:0]   graphic_ops_iterator;
   // 56-bit control word
   // opcode 3 bits wide
   // 3x16 transformation arguments
   // 5 bit knob handler
`define G_OPCODE 55:53
`define G_KNOB 52:48
`define G_XARG 47:32
`define G_YARG 31:16
`define G_ZARG 15:0
   assign      transform_sel[2:0] = graphic_word[`G_OPCODE];
   assign      knob_addr[4:0] = graphic_word[`G_KNOB];
   assign      transarg_x[15:0] = graphic_word[`G_XARG];
   assign      transarg_y[15:0] = graphic_word[`G_YARG];
   assign      transarg_z[15:0] = graphic_word[`G_ZARG];

   signed_mult transop_xer(knobmult_x[17:0], { transarg_x[15:0], 2'b00 }, { knob_val[15:0], 2'b00 });
   signed_mult transop_yer(knobmult_y[17:0], { transarg_y[15:0], 2'b00 }, { knob_val[15:0], 2'b00 });
   signed_mult transop_zer(knobmult_z[17:0], { transarg_z[15:0], 2'b00 }, { knob_val[15:0], 2'b00 });
   assign      transop_x[15:0] = (knob_addr == 0) ? transarg_x[15:0] : knobmult_x[17:2];
   assign      transop_y[15:0] = (knob_addr == 0) ? transarg_y[15:0] : knobmult_y[17:2];
   assign      transop_z[15:0] = (knob_addr == 0) ? transarg_z[15:0] : knobmult_z[17:2];

   rotate rotaterA(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(rotater_init),
                   .busy(rotater_busy_A), .done(rotater_done_A),
                   .Ax(transform_Ax), .Ay(transform_Ay), .Az(transform_Az),
                   .theta(transop_x), .axis(rotate_axis),
                   .Bx(next_Ax), .By(next_Ay), .Bz(next_Az));
   translate translaterA(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(translater_init),
                         .busy(translater_busy_A), .done(translater_done_A),
                         .Ax(transform_Ax), .Ay(transform_Ay), .Az(transform_Az),
                         .Dx(transop_x), .Dy(transop_y), .Dz(transop_z),
                         .Bx(next_Ax), .By(next_Ay), .Bz(next_Az));
   scale scalerA(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(scaler_init),
                 .busy(scaler_busy_A), .done(scaler_done_A),
                 .Ax(transform_Ax), .Ay(transform_Ay), .Az(transform_Az),
                 .Sx(transop_x), .Sy(transop_y), .Sz(transop_z),
                 .Bx(next_Ax), .By(next_Ay), .Bz(next_Az));

   rotate rotaterB(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(rotater_init),
                   .busy(rotater_busy_B), .done(rotater_done_B),
                   .Ax(transform_Bx), .Ay(transform_By), .Az(transform_Bz),
                   .theta(transop_x), .axis(rotate_axis),
                   .Bx(next_Bx), .By(next_By), .Bz(next_Bz));
   translate translaterB(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(translater_init),
                         .busy(translater_busy_B), .done(translater_done_B),
                         .Ax(transform_Bx), .Ay(transform_By), .Az(transform_Bz),
                         .Dx(transop_x), .Dy(transop_y), .Dz(transop_z),
                         .Bx(next_Bx), .By(next_By), .Bz(next_Bz));
   scale scalerB(.CLK(CLK), .reset(MRST), .lock(1'b1), .init(scaler_init),
                 .busy(scaler_busy_B), .done(scaler_done_B),
                 .Ax(transform_Bx), .Ay(transform_By), .Az(transform_Bz),
                 .Sx(transop_x), .Sy(transop_y), .Sz(transop_z),
                 .Bx(next_Bx), .By(next_By), .Bz(next_Bz));

   // This holds reset long enough to clear all the machine state
   initial begin
      MRST = 1;
      #86 MRST = 0;
      #40000
        $finish;
   end

   parameter draw_circle = 4'd0,
               wait_circle = 4'd1,
               draw_lines = 4'd2,
               wait_lines = 4'd3,
               transform_prep = 4'd4,
               transform_exec = 4'd5,
               transform_wait = 4'd6,
               transform_save = 4'd7;
   parameter debug_transform = 1'b0;
   
   // consumer of objects, producer of edge list
   always @(posedge CLK) begin
      if (MRST) begin
         edge_count <= 8'd0;
         object_state <= draw_circle;
      end
      else if (object_state == draw_circle) begin
         //circle_stepsize <= 16'd100;
         //circle_stepsize <= 16'd32;
         circle_stepsize <= 16'd8;
         circle_init <= 1'b1;

         object_state <= wait_circle;
      end
      else if (object_state == wait_circle) begin
         if (circle_busy) begin
            edge_Ax[edge_count] <= circle_Ax;
            edge_Ay[edge_count] <= circle_Ay;
            edge_Az[edge_count] <= circle_Az;
            edge_Bx[edge_count] <= circle_Bx;
            edge_By[edge_count] <= circle_By;
            edge_Bz[edge_count] <= circle_Bz;
            edge_count <= edge_count + 8'd1;

            $display("%d, OBJECT->EDGELIST TICK. init=%b, busy=%b, done=%b, edge_count = %d, [(%h, %h, %h), (%h, %h, %h)] [(%d, %d, %d), (%d, %d, %d)] ",
                     $time, circle_init, circle_busy, circle_done, edge_count,
                     circle_Ax, circle_Ay, circle_Az, circle_Bx, circle_By, circle_Bz,
                     circle_Ax, circle_Ay, circle_Az, circle_Bx, circle_By, circle_Bz);
         end // if (circle_busy)
         else if (circle_done) begin
            $display("circle gen is done. edge_count = %d, edge_iter = %d", edge_count, edge_iterator);
            circle_init <= 1'b0;
         end
      end // if (object_state == wait_circle)
   end // always @ (posedge CLK)

   // consumer of edge list, producer of computed edges
   always @(posedge CLK) begin
      if (MRST) begin
         edge_iterator <= 8'd0;
         comp_count <= 8'd0;

         graphic_ops_iterator <= 0;

         // initialize the graphics ops here
         graphic_ops_count <= 2;
         graphic_ops[0] = { 3'b101, 5'd0, 16'd1 << 13, 16'd1 << 13, 16'd1  << 13};
         //graphic_ops[0] = { 3'b101, 5'd0, 16'd1 << 12, 16'd1 << 12, 16'd1 << 12};
         graphic_ops[1] = { 3'b100, 5'd0, 16'd20, 16'd20, 16'd20 };
         // scale by 0.5
         //graphic_ops[0] = { 3'b101, 5'd0, 16'd1 << 13, 16'd1 << 13, 16'd1  << 13};
         // translate by +20
         //graphic_ops[1] = { 3'b100, 5'd0, 16'd20, 16'd20, 16'd20 };

         // translate +20, +20, +20. works perfectly
         // graphic_word[55:0] = { 3'b100, 5'd0, 16'd20, 16'd20, 16'd20 };
         // scale 2x, 2x, 2x. is broken
         //graphic_word[55:0] = { 3'b101, 5'd0, 16'd2, 16'd2, 16'd2 };
         // for every unit of scale required, must right shift by 14
         //graphic_word[55:0] = { 3'b101, 5'd0, 16'd1 << 13, 16'd1 << 13, 16'd1  << 13};
         // rotate 180 == 256 steps
         //graphic_word[55:0] = { 3'b010, 5'd0, 16'd256, 16'd2, 16'd2 };
         //graphic_word[55:0] = { 3'b010, 5'd0, 16'd128, 16'd0, 16'd0 };

         comp_state <= transform_prep;
      end // if (MRST)
      // blocking on edge_count.
      else if (comp_state == transform_prep) begin
         if (edge_iterator < edge_count) begin
            if (debug_transform) begin
               $display("op dump: %b", graphic_ops[0]);
               $display("op dump: %b", graphic_ops[1]);
            end
            // load transform registers
            transform_Ax <= edge_Ax[edge_iterator] >>> 8;
            transform_Ay <= edge_Ay[edge_iterator] >>> 8;
            transform_Az <= edge_Az[edge_iterator] >>> 8;
            transform_Bx <= edge_Bx[edge_iterator] >>> 8;
            transform_By <= edge_By[edge_iterator] >>> 8;
            transform_Bz <= edge_Bz[edge_iterator] >>> 8;
/*
            transform_Ax <= edge_Ax[edge_iterator];
            transform_Ay <= edge_Ay[edge_iterator];
            transform_Az <= edge_Az[edge_iterator];
            transform_Bx <= edge_Bx[edge_iterator];
            transform_By <= edge_By[edge_iterator];
            transform_Bz <= edge_Bz[edge_iterator];
*/
            // select none
            rotater_init <= 1'b0;
            translater_init <= 1'b0;
            scaler_init <= 1'b0;

            // go to transform_exec
            comp_state <= transform_exec;

            // load graphics op word
            graphic_word <= graphic_ops[graphic_ops_iterator];
         end // if (comp_state == transform_prep)
         else begin
            //$display("in transform_prep waiting for more edges");
         end
      end
      else if (comp_state == transform_exec) begin
         if (graphic_ops_iterator < graphic_ops_count) begin
            if (transform_sel == 3'b000 |
                transform_sel == 3'b001 |
                transform_sel == 3'b010) begin
               // select rotation
               rotater_init <= 1'b1;
               translater_init <= 1'b0;
               scaler_init <= 1'b0;
               if (debug_transform) $display("TRANSFORMATION DISPATCH: op %d of %d: Selecting rotation. theta = %h", graphic_ops_iterator, graphic_ops_count, transop_x);
            end
            else if (transform_sel == 3'b100) begin
               // select translation
               rotater_init <= 1'b0;
               translater_init <= 1'b1;
               scaler_init <= 1'b0;
               if (debug_transform) $display("TRANSFORMATION DISPATCH: op %d of %d: Selecting translation", graphic_ops_iterator, graphic_ops_count);
            end
            else if (transform_sel == 3'b101) begin
               // select scaling
               rotater_init <= 1'b0;
               translater_init <= 1'b0;
               scaler_init <= 1'b1;
               if (debug_transform) $display("TRANSFORMATION DISPATCH: op %d of %d: Selecting scalation", graphic_ops_iterator, graphic_ops_count);
            end
            else begin
               $display("TRANSFORMATION DISPATCH: op %d of %d: wth", graphic_ops_iterator, graphic_ops_count, transop_x);
               $finish;
            end
            
            // wait for transformation to be finished
            //$display("going to transform_wait");
            comp_state <= transform_wait;
            
            // increment to the next operation
            graphic_ops_iterator <= graphic_ops_iterator + 1;
         end // if (graphic_ops_iterator < graphic_ops_count)
         else begin
            $display("transform waiting on exec for more points. graphics_iter = %d, graphics_count = %d", graphic_ops_iterator, graphic_ops_count);
            $finish;
         end
      end // if (comp_state == transform_exec)
      else if (comp_state == transform_wait) begin
         if (rotater_done | translater_done | scaler_done) begin
            if (debug_transform) begin
               $display("Got transformation: From [(%d, %d, %d), (%d, %d, %d)] To [(%d, %d, %d), (%d, %d, %d)]",
                        transform_Ax, transform_Ay, transform_Az, transform_Bx, transform_By, transform_Bz,
                        next_Ax, next_Ay, next_Az, next_Bx, next_By, next_Bz);
            end

            // save the intermediate values
            transform_Ax <= next_Ax;
            transform_Ay <= next_Ay;
            transform_Az <= next_Az;
            transform_Bx <= next_Bx;
            transform_By <= next_By;
            transform_Bz <= next_Bz;
            
            // deselect all the transformation units
            rotater_init <= 1'b0;
            translater_init <= 1'b0;
            scaler_init <= 1'b0;
            
            if (graphic_ops_iterator < graphic_ops_count) begin
               // load graphics op word
               graphic_word <= graphic_ops[graphic_ops_iterator];

               // go back to transform_exec
               comp_state <= transform_exec;
            end
            else begin
               // done state
               comp_state <= transform_save;
            end
         end // if (rotater_done | translater_done | scaler_done)
      end // if (comp_state == transform_wait)
      else if (comp_state == transform_save) begin
         // writeback
         // save point into computed rf
         if (debug_transform) $display("done with ops. incrementing edge_iter from %d. graphic_iter = %d, graphic_count = %d", edge_iterator, graphic_ops_iterator, graphic_ops_count);
         comp_Ax[edge_iterator] <= next_Ax;
         comp_Ay[edge_iterator] <= next_Ay;
         comp_Az[edge_iterator] <= next_Az;
         comp_Bx[edge_iterator] <= next_Bx;
         comp_By[edge_iterator] <= next_By;
         comp_Bz[edge_iterator] <= next_Bz;
         comp_count <= comp_count + 8'd1;
         
         // increment edge iterator
         edge_iterator <= edge_iterator + 8'd1;

         // reset graphics op iterator
         graphic_ops_iterator <= 8'd0;
         
         // go to transform_prep to wait for more edges
         //comp_state <= transform_prep;
         comp_state <= transform_prep;
      end // if (comp_state == transform_save)
   end // always @ (posedge CLK)

   always @(posedge CLK) begin
      if (MRST) begin
         $display("Being reset");

         comp_iterator <= 8'd0;
         raster_state <= draw_lines;
         bresenhem_init <= 1'b0;
      end
      else if (raster_state == draw_lines) begin
         if (comp_iterator < comp_count) begin
            //$display("%d, TICK. init=%b, busy=%b, done=%b, comp_iterator = %d, [(%h, %h, %h), (%h, %h, %h)] [(%d, %d, %d), (%d, %d, %d)] ",
            $display("%d, TICK. init=%b, busy=%b, done=%b, comp_iterator = %d, bresenhem [(%d, %d), (%d, %d)] ",
                     $time, circle_init, circle_busy, circle_done, comp_iterator,
                     //comp_Ax[comp_iterator], comp_Ay[comp_iterator], comp_Az[comp_iterator],
                     //comp_Bx[comp_iterator], comp_By[comp_iterator], comp_Bz[comp_iterator],
                     comp_Ax[comp_iterator] + SCREEN_XHALF, comp_Ay[comp_iterator] + SCREEN_YHALF,
                     comp_Bx[comp_iterator] + SCREEN_XHALF, comp_By[comp_iterator] + SCREEN_YHALF);
                     //comp_Ax[comp_iterator], comp_Ay[comp_iterator], comp_Az[comp_iterator],
                     //comp_Bx[comp_iterator], comp_By[comp_iterator], comp_Bz[comp_iterator]);
            //(comp_Ax[comp_iterator] >>> 8) + SCREEN_XHALF, (comp_Ay[comp_iterator] >>> 8) + SCREEN_YHALF, comp_Az[comp_iterator] >>> 8,
            //(comp_Bx[comp_iterator] >>> 8) + SCREEN_XHALF, (comp_By[comp_iterator] >>> 8) + SCREEN_YHALF, comp_Bz[comp_iterator] >>> 8);

            bre_Ax <= comp_Ax[comp_iterator] + SCREEN_XHALF;
            bre_Ay <= comp_Ay[comp_iterator] + SCREEN_YHALF;
            bre_Bx <= comp_Bx[comp_iterator] + SCREEN_XHALF;
            bre_By <= comp_By[comp_iterator] + SCREEN_YHALF;
            bresenhem_init <= 1'b1;
            
            comp_iterator <= comp_iterator + 8'd1;
            
         
            raster_state <= wait_lines;
         end // if (comp_iterator < comp_count)
         else begin
            //$display("raster is waiting for more points");
         end
      end // if (raster_state <= draw_lines)
      else if (raster_state == wait_lines) begin
         if (bresenhem_busy) begin
            $display("Waiting for bresenhem");
         end
         else if (bresenhem_done) begin
            $display("bresenhem is done");
            bresenhem_init <= 1'b0;
            raster_state <= draw_lines;
         end
         else begin
            //$display("at the end of wait_lines");
         end
      end // if (raster_state == wait_lines)
      else begin
         //$display("ererefalsdkfslkjfklj");
      end
   end // always @ (posedge CLK)

endmodule // TOP
